@@ -12,6 +12,7 @@
 #include "../libpcsxcore/misc.h"
 #include "../libpcsxcore/psemu_plugin_defs.h"
 #include "../libpcsxcore/cdrom.h"
+#include "../libpcsxcore/cdriso.h"
 #include "common/readpng.h"
 #include "maemo_common.h"
 #include <libosso.h>
@@ -32,6 +33,7 @@
 
 osso_context_t* osso;
 extern int bKeepDisplayOn;
+extern int bAutosaveOnExit;
 extern char keys_config_file[MAXPATHLEN];
 static pthread_t display_thread = (pthread_t)-1;
 int g_layer_x = (X_RES - D_WIDTH) / 2;
@@ -81,11 +83,11 @@ window_key_proxy(GtkWidget *widget,
 static void *displayThread(void *arg)
 {
 	DBusConnection* system_bus = (DBusConnection*)osso_get_sys_dbus_connection(osso);
-	DBusMessage* msg = dbus_message_new_method_call("com.nokia.mce", 
+	DBusMessage* msg = dbus_message_new_method_call("com.nokia.mce",
 												    "/com/nokia/mce/request",
-												    "com.nokia.mce.request", 
+												    "com.nokia.mce.request",
 												    "req_display_blanking_pause");
-	if (msg && system_bus) { 
+	if (msg && system_bus) {
 		while (1) {
 			dbus_connection_send(system_bus, msg, NULL);
 			dbus_connection_flush(system_bus);
@@ -98,13 +100,15 @@ static void *displayThread(void *arg)
 
 void show_notification(char* text)
 {
-	DBusConnection* session_bus = (DBusConnection*)osso_get_dbus_connection(osso);
+    GtkWidget* banner = hildon_banner_show_information(GTK_WIDGET(window), NULL, text);
+    hildon_banner_set_timeout(HILDON_BANNER(banner), 3000);
+	/*DBusConnection* session_bus = (DBusConnection*)osso_get_dbus_connection(osso);
 	DBusMessageIter args;
-	DBusMessage*msg = dbus_message_new_method_call("org.freedesktop.Notifications", 
+	DBusMessage*msg = dbus_message_new_method_call("org.freedesktop.Notifications",
 												   "/org/freedesktop/Notifications",
-												   "org.freedesktop.Notifications", 
+												   "org.freedesktop.Notifications",
 												   "SystemNoteInfoprint");
-	if (msg) { 
+	if (msg) {
 		dbus_message_iter_init_append(msg, &args);
 		char* param = text;
 		if (dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &param)) {
@@ -112,7 +116,7 @@ void show_notification(char* text)
 			dbus_connection_flush(session_bus);
 		}
 		dbus_message_unref(msg);
-	}
+	}*/
 }
 
 void show_messagebox(char* text)
@@ -142,11 +146,17 @@ void change_disc()
 
 	GtkFileFilter *filter=gtk_file_filter_new();
 	gtk_file_filter_add_pattern (filter,"*.bin");
+	gtk_file_filter_add_pattern (filter,"*.BIN");
 	gtk_file_filter_add_pattern (filter,"*.iso");
+	gtk_file_filter_add_pattern (filter,"*.ISO");
 	gtk_file_filter_add_pattern (filter,"*.img");
+	gtk_file_filter_add_pattern (filter,"*.IMG");
 	gtk_file_filter_add_pattern (filter,"*.z");
+	gtk_file_filter_add_pattern (filter,"*.Z");
 	gtk_file_filter_add_pattern (filter,"*.znx");
+	gtk_file_filter_add_pattern (filter,"*.ZNX");
 	gtk_file_filter_add_pattern (filter,"*.pbp");
+	gtk_file_filter_add_pattern (filter,"*.PBP");
 	gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog),filter);
 
 	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK) {
@@ -159,9 +169,9 @@ void change_disc()
 			set_cd_image(filename);
 			if (ReloadCdromPlugin() < 0)
 				printf("Failed to load cdr plugin\n");
-			
+
 			if (CDR_open() < 0)
-				printf("Failed to open cdr plugin\n");		
+				printf("Failed to open cdr plugin\n");
 
 			strcpy(file_name, strrchr(filename,'/'));
 
@@ -174,10 +184,91 @@ void change_disc()
 	gtk_widget_destroy (dialog);
 }
 
+void change_multi_disc()
+{
+    HildonDialog* window = HILDON_DIALOG(hildon_dialog_new());
+    gtk_window_set_title (GTK_WINDOW (window), "Change disc");
+    gtk_window_set_default_size(GTK_WINDOW (window), 480, 300);
+
+    GtkWidget* sw = hildon_pannable_area_new ();
+    gtk_box_pack_start (GTK_BOX(GTK_DIALOG(window)->vbox), sw, TRUE, TRUE, 0);
+
+    GtkWidget* tree_view = hildon_gtk_tree_view_new (HILDON_UI_MODE_EDIT);
+    gtk_widget_set_name (tree_view, "fremantle-widget");
+
+    gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (tree_view), TRUE);
+
+    int i;
+    GtkListStore *store = gtk_list_store_new (1, G_TYPE_STRING);
+    for (i = 0; i < cdrIsoMultidiskCount; i++) {
+        gchar *str;
+
+        str = g_strdup_printf ("Disc %d", i+1);
+        gtk_list_store_insert_with_values (store, NULL, i, 0, str, -1);
+        g_free (str);
+    }
+    GtkTreeModel* model =  GTK_TREE_MODEL (store);
+
+    gtk_tree_view_set_model (GTK_TREE_VIEW (tree_view), model);
+    g_object_unref (model);
+
+    GtkTreeSelection* selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree_view));
+    gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
+
+    GtkCellRenderer* renderer = gtk_cell_renderer_text_new ();
+    g_object_set (renderer,
+                  "xalign", 0.5,
+                  "weight", PANGO_WEIGHT_NORMAL,
+                  NULL);
+
+    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (tree_view),
+                                                 0, "Column 0",
+                                                 renderer,
+                                                 "text", 0,
+                                                 NULL);
+
+    char current[5];
+    sprintf(current, "%i", cdrIsoMultidiskSelect);
+    GtkTreePath* path = gtk_tree_path_new_from_string(current);
+    gtk_tree_selection_select_path (selection, path);
+    gtk_tree_path_free(path);
+
+    gtk_widget_set_size_request (tree_view, 480, 800);
+    gtk_container_add (GTK_CONTAINER (sw), tree_view);
+
+    hildon_dialog_add_button (HILDON_DIALOG(window), GTK_STOCK_OK, GTK_RESPONSE_ACCEPT);
+
+    gtk_widget_show_all (GTK_WIDGET(window));
+    gint result = gtk_dialog_run (GTK_DIALOG (window));
+    if (result == GTK_RESPONSE_ACCEPT) {
+      GtkTreeModel* model;
+      GtkTreeIter iter;
+      GtkTreeSelection* selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
+      if (gtk_tree_selection_get_selected(selection, &model, &iter)){
+	    GtkTreePath* path = gtk_tree_model_get_path(model , &iter);
+		int* i = gtk_tree_path_get_indices(path) ;
+
+		cdrIsoMultidiskSelect = *i;
+		CdromId[0] = '\0';
+		CdromLabel[0] = '\0';
+
+		CDR_close();
+		if (CDR_open() < 0) {
+			printf("Failed to load cdr plugin\n");
+			return;
+		}
+
+		SetCdOpenCaseTime(time(NULL) + 3);
+		LidInterrupt();
+      }
+    }
+	gtk_widget_destroy(GTK_WIDGET(window));
+}
+
 static DBusHandlerResult on_msg_recieved(DBusConnection* connection G_GNUC_UNUSED, DBusMessage* message, void* data)
 {
-	const char* msg = dbus_message_get_path(message);
-	if (msg && g_str_equal(msg, CALL_SIGNAL_PATH)){
+	const char* path = dbus_message_get_path(message);
+	if (path && g_str_equal(path, CALL_SIGNAL_PATH)){
 		const char* mbr = dbus_message_get_member(message);
 		if (mbr && g_str_equal(mbr, CALL_INCOMING_SIG))
 			show_messagebox("Paused");
@@ -199,6 +290,15 @@ inline void key_press_event(int key2,int type)
 		psxkey1 = keymap2[key];
 	else switch (key) {
 		case 14:
+			if (bAutosaveOnExit){
+			    show_notification("Autosaving");
+				emu_save_state(99);
+				char buf[MAXPATHLEN];
+				if (image && image->mem){
+                    sprintf (buf,"/opt/maemo/usr/games/screenshots%s.%3.3d",file_name,99);
+                    writepng(buf, image->mem, pl_buf_w,pl_buf_h);
+				}
+			}
 			hildon_quit();
 			break;
 		case 15:
@@ -252,8 +352,12 @@ inline void key_press_event(int key2,int type)
 			}
 			return;
 		case 23:
-			if (type == 1)
-				change_disc();
+			if (type == 1){
+				if (cdrIsoMultidiskCount > 1)
+					change_multi_disc();
+				else
+					change_disc();
+			}
 			return;
 	}
 
@@ -295,7 +399,7 @@ int maemo_init(int *argc, char ***argv)
 	FILE* pFile;
 	pFile = fopen(keys_config_file, "r");
 	if (pFile == NULL){
-		printf("Error opening keys config file %s\n", keys_config_file);
+		fprintf(stderr, "Error opening keys config file %s\n", keys_config_file);
 		return 1;
 	}
 	printf("Keys config read from %s\n", keys_config_file);
@@ -354,8 +458,10 @@ int maemo_init(int *argc, char ***argv)
 
 void maemo_finish()
 {
-    osso_deinitialize(osso);
-	osso = NULL;
+	if (osso){
+		osso_deinitialize(osso);
+		osso = NULL;
+	}
 }
 
 void menu_loop(void)
