@@ -1,4 +1,5 @@
-void maemo_init(int *argc, char ***argv);
+int maemo_init(int *argc, char ***argv);
+void maemo_finish();
 
 extern char file_name[MAXPATHLEN];
 extern int g_maemo_opts;
