@@ -15,6 +15,7 @@
 #include "plugin.h"
 #include "plugin_lib.h"
 #include "../libpcsxcore/misc.h"
+#include "../libpcsxcore/cdriso.h"
 #include "../libpcsxcore/new_dynarec/new_dynarec.h"
 #include "../plugins/dfinput/main.h"
 #include "maemo_common.h"
@@ -31,6 +32,7 @@ accel_option accelOptions;
 int g_opts = 0;
 int g_maemo_opts;
 int bKeepDisplayOn = FALSE;
+int bAutosaveOnExit = FALSE;
 char file_name[MAXPATHLEN];
 char keys_config_file[MAXPATHLEN] = "/opt/psx4m/keys";
 
@@ -58,12 +60,14 @@ void PrintHelp()
 
 	printf("Options:\n");
 	printf("  -help          : This help\n");
+	printf("  -disc VALUE    : Disc number for multi discs images\n");
 	printf("  -fullscreen    : Run fullscreen\n");
 	printf("  -frameskip     : Frameskip\n");
 	printf("                   -1=Auto (Default)\n");
 	printf("                    0=Disabled\n");
 	printf("                    1=Set to 1\n");
 	printf("                    ...\n");
+ 	printf("  -autosave      : Enable auto save on exit\n");
 	printf("  -accel         : Enable accelerometer\n");
 	printf("  -analog        : Use analog pad for accel\n");
 	printf("  -sens VALUE    : Set accelerometer sens [0-1000]\n");
@@ -166,6 +170,7 @@ int maemo_main(int argc, char **argv)
 	int loadst = 0;
 	int i;
 	int getst = -1;
+	int discNumber = 0;
 
 	strcpy(Config.Gpu, "builtin_gpu");
 	strcpy(Config.Spu, "builtin_spu");
@@ -252,6 +257,8 @@ int maemo_main(int argc, char **argv)
 		else if (!strcmp(argv[i], "-max")) 				    accelOptions.maxValue = atol(argv[++i]);
 		else if (!strcmp(argv[i], "-displayon")) 		    bKeepDisplayOn = TRUE;
 		else if (!strcmp(argv[i], "-keys")) 				sprintf(keys_config_file, "%s", argv[++i]);
+		else if (!strcmp(argv[i], "-autosave")) 		    bAutosaveOnExit = TRUE;
+		else if (!strcmp(argv[i], "-disc")) 		        discNumber = atol(argv[++i]);
 
 		else if (!strcmp(argv[i], "-spu_reverb"))           { if (atol(argv[++i]) > 0) iUseReverb = 2; }
 		else if (!strcmp(argv[i], "-spu_interpolation")) 	iUseInterpolation = atol(argv[++i]);
@@ -269,7 +276,7 @@ int maemo_main(int argc, char **argv)
         else if (!strcmp(argv[i], "-gles_opaque")) 	    pl_rearmed_cbs.gpu_peopsgl.bOpaquePass = atol(argv[++i]);
 
 		else {
-			printf("Unknown option: %s\n", argv[i]);
+			fprintf(stderr, "Unknown option: %s\n", argv[i]);
 			return 1;
 		}
 	}
@@ -291,6 +298,9 @@ int maemo_main(int argc, char **argv)
 		return 1;
 	}
 
+	if (discNumber > 0)
+		cdrIsoMultidiskSelect = discNumber - 1;
+
 	if (OpenPlugins() == -1) {
 		return 1;
 	}
@@ -303,6 +313,21 @@ int maemo_main(int argc, char **argv)
 
 		get_state_filename(fname, sizeof(fname), getst);
 		printf("SAVESTATE: %s\n", fname);
+		if (cdrIsoMultidiskCount > 1){
+			int i = 0;
+			for (i=1; i<cdrIsoMultidiskCount; i++){
+				cdrIsoMultidiskSelect = i;
+				CdromId[0] = '\0';
+				CdromLabel[0] = '\0';
+
+				CDR_close();
+				if (CDR_open() == 0){
+					CheckCdrom();
+					get_state_filename(fname, sizeof(fname), getst);
+					printf("SAVESTATE: %s\n", fname);
+				}
+			}
+		}
 		return 0;
 	}
 
@@ -327,6 +352,9 @@ int maemo_main(int argc, char **argv)
 		return 1;
 	}
 
+	if (cdrIsoMultidiskCount > 1)
+		printf ("Loaded a multidisc image: %i discs.\n", cdrIsoMultidiskCount);
+
 	// If a state has been specified, then load that
 	if (loadst) {
 		int ret = emu_load_state(loadst - 1);
