@@ -21,8 +21,6 @@
 
 #include "externals.h"
 #include "registers.h"
-#include "regs.h"
-#include "reverb.h"
 
 /*
 // adsr time values (in ms) by James Higgs ... see the end of
@@ -42,6 +40,15 @@
 #define SUSTAIN_MS     441L
 #define RELEASE_MS     437L
 
+static void SoundOn(int start,int end,unsigned short val);
+static void SoundOff(int start,int end,unsigned short val);
+static void FModOn(int start,int end,unsigned short val);
+static void NoiseOn(int start,int end,unsigned short val);
+static void SetVolumeL(unsigned char ch,short vol);
+static void SetVolumeR(unsigned char ch,short vol);
+static void SetPitch(int ch,unsigned short val);
+static void ReverbOn(int start,int end,unsigned short val);
+
 ////////////////////////////////////////////////////////////////////////
 // WRITE REGISTERS: called by main emu
 ////////////////////////////////////////////////////////////////////////
@@ -70,8 +77,7 @@ void CALLBACK SPUwriteRegister(unsigned long reg, unsigned short val)
        break;
      //------------------------------------------------// start
      case 6:      
-       // Brain Dead 13 - align to 16 boundary
-       s_chan[ch].pStart= spuMemC+(unsigned long)((val<<3)&~0xf);
+       // taken from regArea later
        break;
      //------------------------------------------------// level with pre-calcs
      case 8:
@@ -166,10 +172,11 @@ void CALLBACK SPUwriteRegister(unsigned long reg, unsigned short val)
        break;
      //------------------------------------------------//
      case 14:                                          // loop?
-       //WaitForSingleObject(s_chan[ch].hMutex,2000);        // -> no multithread fuckups
-       s_chan[ch].pLoop=spuMemC+((unsigned long)((val<<3)&~0xf));
-       s_chan[ch].bIgnoreLoop=1;
-       //ReleaseMutex(s_chan[ch].hMutex);                    // -> oki, on with the thread
+       s_chan[ch].pLoop=spuMemC+((val&~1)<<3);
+       if(s_chan[ch].bJump)
+        // real machine would be most likely still doing the last block and use new value for the jump;
+        // but we decode ahead a bit and already did the jump part, so compensate for that now.
+        s_chan[ch].pCurr=s_chan[ch].pLoop;
        break;
      //------------------------------------------------//
     }
@@ -191,6 +198,8 @@ void CALLBACK SPUwriteRegister(unsigned long reg, unsigned short val)
       break;
     //-------------------------------------------------//
     case H_SPUctrl:
+      if(!(spuCtrl & CTRL_IRQ))
+        spuStat&=~STAT_IRQ;
       spuCtrl=val;
       break;
     //-------------------------------------------------//
@@ -210,11 +219,12 @@ void CALLBACK SPUwriteRegister(unsigned long reg, unsigned short val)
           rvb.CurrAddr=rvb.StartAddr;
          }
        }
+      rvb.dirty = 1;
       break;
     //-------------------------------------------------//
     case H_SPUirqAddr:
       spuIrq = val;
-      pSpuIrq=spuMemC+((unsigned long) val<<3);
+      pSpuIrq=spuMemC+(((unsigned long) val<<3)&~0xf);
       break;
     //-------------------------------------------------//
     case H_SPUrvolL:
@@ -301,19 +311,8 @@ void CALLBACK SPUwriteRegister(unsigned long reg, unsigned short val)
       ReverbOn(16,24,val);
       break;
     //-------------------------------------------------//
-    case H_Reverb+0:
-
-      rvb.FB_SRC_A=val;
-
-      // OK, here's the fake REVERB stuff...
-      // depending on effect we do more or less delay and repeats... bah
-      // still... better than nothing :)
-
-      SetREVERB(val);
-      break;
-
-
-    case H_Reverb+2   : rvb.FB_SRC_B=(short)val;       break;
+    case H_Reverb+0   : rvb.FB_SRC_A=val*4;            break;
+    case H_Reverb+2   : rvb.FB_SRC_B=val*4;            break;
     case H_Reverb+4   : rvb.IIR_ALPHA=(short)val;      break;
     case H_Reverb+6   : rvb.ACC_COEF_A=(short)val;     break;
     case H_Reverb+8   : rvb.ACC_COEF_B=(short)val;     break;
@@ -322,30 +321,33 @@ void CALLBACK SPUwriteRegister(unsigned long reg, unsigned short val)
     case H_Reverb+14  : rvb.IIR_COEF=(short)val;       break;
     case H_Reverb+16  : rvb.FB_ALPHA=(short)val;       break;
     case H_Reverb+18  : rvb.FB_X=(short)val;           break;
-    case H_Reverb+20  : rvb.IIR_DEST_A0=(short)val;    break;
-    case H_Reverb+22  : rvb.IIR_DEST_A1=(short)val;    break;
-    case H_Reverb+24  : rvb.ACC_SRC_A0=(short)val;     break;
-    case H_Reverb+26  : rvb.ACC_SRC_A1=(short)val;     break;
-    case H_Reverb+28  : rvb.ACC_SRC_B0=(short)val;     break;
-    case H_Reverb+30  : rvb.ACC_SRC_B1=(short)val;     break;
-    case H_Reverb+32  : rvb.IIR_SRC_A0=(short)val;     break;
-    case H_Reverb+34  : rvb.IIR_SRC_A1=(short)val;     break;
-    case H_Reverb+36  : rvb.IIR_DEST_B0=(short)val;    break;
-    case H_Reverb+38  : rvb.IIR_DEST_B1=(short)val;    break;
-    case H_Reverb+40  : rvb.ACC_SRC_C0=(short)val;     break;
-    case H_Reverb+42  : rvb.ACC_SRC_C1=(short)val;     break;
-    case H_Reverb+44  : rvb.ACC_SRC_D0=(short)val;     break;
-    case H_Reverb+46  : rvb.ACC_SRC_D1=(short)val;     break;
-    case H_Reverb+48  : rvb.IIR_SRC_B1=(short)val;     break;
-    case H_Reverb+50  : rvb.IIR_SRC_B0=(short)val;     break;
-    case H_Reverb+52  : rvb.MIX_DEST_A0=(short)val;    break;
-    case H_Reverb+54  : rvb.MIX_DEST_A1=(short)val;    break;
-    case H_Reverb+56  : rvb.MIX_DEST_B0=(short)val;    break;
-    case H_Reverb+58  : rvb.MIX_DEST_B1=(short)val;    break;
+    case H_Reverb+20  : rvb.IIR_DEST_A0=val*4;         break;
+    case H_Reverb+22  : rvb.IIR_DEST_A1=val*4;         break;
+    case H_Reverb+24  : rvb.ACC_SRC_A0=val*4;          break;
+    case H_Reverb+26  : rvb.ACC_SRC_A1=val*4;          break;
+    case H_Reverb+28  : rvb.ACC_SRC_B0=val*4;          break;
+    case H_Reverb+30  : rvb.ACC_SRC_B1=val*4;          break;
+    case H_Reverb+32  : rvb.IIR_SRC_A0=val*4;          break;
+    case H_Reverb+34  : rvb.IIR_SRC_A1=val*4;          break;
+    case H_Reverb+36  : rvb.IIR_DEST_B0=val*4;         break;
+    case H_Reverb+38  : rvb.IIR_DEST_B1=val*4;         break;
+    case H_Reverb+40  : rvb.ACC_SRC_C0=val*4;          break;
+    case H_Reverb+42  : rvb.ACC_SRC_C1=val*4;          break;
+    case H_Reverb+44  : rvb.ACC_SRC_D0=val*4;          break;
+    case H_Reverb+46  : rvb.ACC_SRC_D1=val*4;          break;
+    case H_Reverb+48  : rvb.IIR_SRC_B1=val*4;          break;
+    case H_Reverb+50  : rvb.IIR_SRC_B0=val*4;          break;
+    case H_Reverb+52  : rvb.MIX_DEST_A0=val*4;         break;
+    case H_Reverb+54  : rvb.MIX_DEST_A1=val*4;         break;
+    case H_Reverb+56  : rvb.MIX_DEST_B0=val*4;         break;
+    case H_Reverb+58  : rvb.MIX_DEST_B1=val*4;         break;
     case H_Reverb+60  : rvb.IN_COEF_L=(short)val;      break;
     case H_Reverb+62  : rvb.IN_COEF_R=(short)val;      break;
    }
 
+ if ((r & ~0x3f) == H_Reverb)
+  rvb.dirty = 1; // recalculate on next update
+
  iSpuAsyncWait=0;
 }
 
@@ -376,7 +378,6 @@ unsigned short CALLBACK SPUreadRegister(unsigned long reg)
      case 14:                                          // get loop address
       {
        const int ch=(r>>4)-0xc0;
-       if(s_chan[ch].pLoop==NULL) return 0;
        return (unsigned short)((s_chan[ch].pLoop-spuMemC)>>3);
       }
     }
@@ -419,23 +420,24 @@ unsigned short CALLBACK SPUreadRegister(unsigned long reg)
 // SOUND ON register write
 ////////////////////////////////////////////////////////////////////////
 
-void SoundOn(int start,int end,unsigned short val)     // SOUND ON PSX COMAND
+static void SoundOn(int start,int end,unsigned short val)
 {
  int ch;
 
  for(ch=start;ch<end;ch++,val>>=1)                     // loop channels
   {
-   if((val&1) && s_chan[ch].pStart)                    // mmm... start has to be set before key on !?!
+   if((val&1) && regAreaGet(ch,6))                     // mmm... start has to be set before key on !?!
     {
-     s_chan[ch].bIgnoreLoop=0;
-
      // do this here, not in StartSound
      // - fixes fussy timing issues
      s_chan[ch].bStop=0;
-     s_chan[ch].pCurr=s_chan[ch].pStart;
+     s_chan[ch].pCurr=spuMemC+((regAreaGet(ch,6)&~1)<<3); // must be block aligned
+     s_chan[ch].pLoop=spuMemC+((regAreaGet(ch,14)&~1)<<3);
+     s_chan[ch].bJump=0;
 
      dwNewChannel|=(1<<ch);                            // bitfield for faster testing
      dwChannelOn|=1<<ch;
+     dwChannelDead&=~(1<<ch);
     }
   }
 }
@@ -444,7 +446,7 @@ void SoundOn(int start,int end,unsigned short val)     // SOUND ON PSX COMAND
 // SOUND OFF register write
 ////////////////////////////////////////////////////////////////////////
 
-void SoundOff(int start,int end,unsigned short val)    // SOUND OFF PSX COMMAND
+static void SoundOff(int start,int end,unsigned short val)
 {
  int ch;
  for(ch=start;ch<end;ch++,val>>=1)                     // loop channels
@@ -464,7 +466,7 @@ void SoundOff(int start,int end,unsigned short val)    // SOUND OFF PSX COMMAND
 // FMOD register write
 ////////////////////////////////////////////////////////////////////////
 
-void FModOn(int start,int end,unsigned short val)      // FMOD ON PSX COMMAND
+static void FModOn(int start,int end,unsigned short val)
 {
  int ch;
 
@@ -481,6 +483,8 @@ void FModOn(int start,int end,unsigned short val)      // FMOD ON PSX COMMAND
    else
     {
      s_chan[ch].bFMod=0;                               // --> turn off fmod
+     if(ch>0&&s_chan[ch-1].bFMod==2)
+      s_chan[ch-1].bFMod=0;
     }
   }
 }
@@ -489,7 +493,7 @@ void FModOn(int start,int end,unsigned short val)      // FMOD ON PSX COMMAND
 // NOISE register write
 ////////////////////////////////////////////////////////////////////////
 
-void NoiseOn(int start,int end,unsigned short val)     // NOISE ON PSX COMMAND
+static void NoiseOn(int start,int end,unsigned short val)
 {
  int ch;
 
@@ -506,7 +510,7 @@ void NoiseOn(int start,int end,unsigned short val)     // NOISE ON PSX COMMAND
 // please note: sweep and phase invert are wrong... but I've never seen
 // them used
 
-void SetVolumeL(unsigned char ch,short vol)            // LEFT VOLUME
+static void SetVolumeL(unsigned char ch,short vol)     // LEFT VOLUME
 {
  if(vol&0x8000)                                        // sweep?
   {
@@ -532,7 +536,7 @@ void SetVolumeL(unsigned char ch,short vol)            // LEFT VOLUME
 // RIGHT VOLUME register write
 ////////////////////////////////////////////////////////////////////////
 
-void SetVolumeR(unsigned char ch,short vol)            // RIGHT VOLUME
+static void SetVolumeR(unsigned char ch,short vol)     // RIGHT VOLUME
 {
  if(vol&0x8000)                                        // comments... see above :)
   {
@@ -558,24 +562,22 @@ void SetVolumeR(unsigned char ch,short vol)            // RIGHT VOLUME
 // PITCH register write
 ////////////////////////////////////////////////////////////////////////
 
-void SetPitch(int ch,unsigned short val)               // SET PITCH
+static void SetPitch(int ch,unsigned short val)               // SET PITCH
 {
  int NP;
  if(val>0x3fff) NP=0x3fff;                             // get pitch val
  else           NP=val;
 
  s_chan[ch].iRawPitch=NP;
-
- NP=(44100L*NP)/4096L;                                 // calc frequency
- if(NP<1) NP=1;                                        // some security
- s_chan[ch].iActFreq=NP;                               // store frequency
+ s_chan[ch].sinc=(NP<<4)|8;
+ if(iUseInterpolation==1) s_chan[ch].SB[32]=1;         // -> freq change in simple interpolation mode: set flag
 }
 
 ////////////////////////////////////////////////////////////////////////
 // REVERB register write
 ////////////////////////////////////////////////////////////////////////
 
-void ReverbOn(int start,int end,unsigned short val)    // REVERB ON PSX COMMAND
+static void ReverbOn(int start,int end,unsigned short val)
 {
  int ch;
 
